<?php

namespace Ulmus\Ldap\Common;

use http\Exception\InvalidArgumentException;
use LDAP\Result;
use function ldap_set_option, ldap_start_tls, ldap_bind_ext, ldap_unbind, ldap_connect, ldap_close, ldap_get_entries, ldap_mod_replace, ldap_count_entries, ldap_errno, ldap_error;

class LdapObject {

    public static ? string $dump = null;

    public $connection;

    public $search;

    protected string $host;

    public int $rowCount = 0;

    public int $bufferedRows = 0;

    public /*int*/ $lastInsertId = 0;

    public string $dn;

    public false|Result $binded = false;

    public false|Result $lastQuery = false;

    public bool $deleteRecursively = true;

    public function __destruct()
    {
        isset($this->connection) and ldap_close($this->connection);
    }

    public function connect(string $host, string $baseDn) : bool
    {
        $this->host = $host;

        $this->connection = ldap_connect("ldap://$host");

        $this->dn = $baseDn;

        return $this->connection !== false;
    }

    public function setOptions(array $options, $useConnectionRessource = true) : void
    {
        foreach($options as $field => $value) {
            ldap_set_option($useConnectionRessource ? $this->connection : null, $field, $value);
        }
    }

    public function startTLS() : void
    {
        ldap_start_tls($this->connection);
    }

    public function bind(? string $dn, ? string $password = null) : false|Result
    {
        if ($this->binded) {
            throw new \Exception("LdapObject is already binded with a user. Use the unbind() method to release it.");
        }

        try {
            $this->binded = ldap_bind_ext($this->connection, $dn, $password);
        }
        catch(\Throwable $ex) {
            throw new \ErrorException(sprintf("%s [ using: %s on %s ]", $ex->getMessage(), $dn, $this->host ));
        }

        return $this->binded;
    }

    public function unbind() : void
    {
        if ($this->binded) {
            $this->binded = ! ldap_unbind($this->connection);
        }
    }

    public function rebind(? string $dn, ? string $password = null) : bool
    {
        $this->unbind();

        return $this->bind($dn, $password);
    }

    public function compare(? string $dn, string $attribute, /* mixed */ $value) /*: bool|int*/
    {
        return ldap_compare($this->connection, $dn, $attribute, $value);
    }

    public function select(array $filter, array $fields = [])
    {
        static::$dump && call_user_func_array(static::$dump, [ $filter, $fields ]);

        $this->search = @ldap_search($this->connection, $this->dn, $filter['filters'], $filter['fields'] ?? [], 0, $filter['limit'] ?? 0);

        $this->throwLdapException();

        $this->rowCount = $this->bufferedRows = ldap_count_entries($this->connection, $this->search);

        return $this;
    }

    public function fetch() /* : bool|array */
    {
        static $result;

        if (! $this->search ) {
            throw new \Exception('Impossible to fetch from LdapObject from which select() was not called first.');
        }

        if ( ! $this->bufferedRows ) {
            return $result = false;
        }
        if ( $this->rowCount === $this->bufferedRows ) {
            $result = ldap_first_entry($this->connection, $this->search);
        }
        else {
            $result = ldap_next_entry($this->connection, $result);
        }

        if ($result) {
            $this->bufferedRows--;

            $dataset = [
                'dn' => ldap_get_dn($this->connection, $result)
            ];

            if ( $attributes = ldap_get_attributes($this->connection, $result) ) {
                for ($i = 0; $i < $attributes['count']; $i++) {
                    $key = $attributes[$i];
                    $dataset[$key] = $attributes[$key]['count'] > 1 ? array_diff_key($attributes[$key], [ 'count' => "" ]) : $attributes[$key][0];
                }
            }

            return $dataset;
        }

        return false;
    }

    #public function fetchAll()
    #{
    #    return ldap_get_entries($this->connection, $result);
    #}

    public function rowCount() : int
    {
        return $this->rowCount;
    }

    /**
     * @deprecated
     */
    public function runQuery(array $filter, array $dataset)
    {
        throw new \Exception("Method runQuery() cannot be run on this connector");
    }

    public function runInsertQuery(array $filter, array $dataset)
    {
        static::$dump && call_user_func_array(static::$dump, [ $filter, $dataset ]);

        $dn = $filter['dn'] ?? $this->dn;

        $combine = array_combine($filter['fields'], $dataset);

        if ( $combine['cn'] ?? false ) {
            $dn = "cn={$combine['cn']}," . $dn;
        }

        try {

            $this->lastQuery = ldap_add_ext($this->connection, $dn, $combine);

            if ($this->lastQuery === false) {
                $this->throwLdapException();
            }
            else {
                $this->rowCount = 1;
                $this->lastInsertId = $dn;
            }
        }
        catch(\Throwable $e) {
            $this->throwLdapException();
        }

        return $this;
    }

    public function runAddQuery(array $filter, array $dataset)
    {
        static::$dump && call_user_func_array(static::$dump, [ $filter, $dataset ]);

        $this->lastQuery = ldap_mod_add_ext($this->connection, $filter['dn'], $dataset);

        if ( $this->lastQuery === false ) {
            $this->throwLdapException();
        }

        $this->rowCount = $this->lastQuery === false ? 0 : 1;

        return $this;
    }

    public function runUpdateQuery(array $filter, array $dataset)
    {
        static::$dump && call_user_func_array(static::$dump, [ $filter, $dataset ]);

        $this->lastQuery = ldap_mod_replace_ext($this->connection, $filter['dn'], $dataset);

        if ( $this->lastQuery === false ) {
            $this->throwLdapException();
        }

        $this->rowCount = $this->lastQuery === false ? 0 : 1;

        return $this;
    }

    public function runDeleteQuery(array $filter)
    {
        static::$dump && call_user_func_array(static::$dump, [ $filter ]);

        if ( empty($filter['dn']) ) {
            throw new InvalidArgumentException("A valid DN must be provided to run a 'delete' query on LDAP connector");
        }

        if ($this->deleteRecursively) {
            $list = ldap_list($this->connection, $filter['dn'], "ObjectClass=*", [""]);
            $info = ldap_get_entries($this->connection, $list);

            for($i=0; $i < $info['count']; $i++){
                $this->runDeleteQuery($info[$i]);
            }
        }

        $this->lastQuery = ldap_delete_ext($this->connection, $filter['dn']);

        if ( $this->lastQuery === false ) {
            $this->throwLdapException();
        }

        $this->rowCount = $this->lastQuery === false ? 0 : 1;

        return $this;
    }

    protected function throwLdapException() : void
    {
        $no = ldap_errno($this->connection);

        switch($no) {
            case 0: # Success
            case 4: # Skipping 'Size limit exceeded' error
                return;
        }

        ldap_get_option($this->connection, \LDAP_OPT_DIAGNOSTIC_MESSAGE, $err);

        throw new \Exception(sprintf('LDAP error #%s `%s`. %s', $no, ldap_error($this->connection), $err));
    }

    public function closeCursor() : void {}
}
